package be.iminds.ilabt.jfed.highlevel.model;

import be.iminds.ilabt.jfed.log.ApiCallDetails;
import be.iminds.ilabt.jfed.lowlevel.AnyCredential;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.api.AbstractFederationApi;
import be.iminds.ilabt.jfed.lowlevel.api.FederationMemberAuthorityApi1;
import be.iminds.ilabt.jfed.lowlevel.api.FederationMemberAuthorityApi2;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.KeyUtil;
import javafx.application.Platform;
import org.slf4j.LoggerFactory;

import java.security.cert.CertificateParsingException;
import java.security.cert.X509Certificate;
import java.util.Collection;
import java.util.List;
import java.util.Map;

/**
 * SfaModelSliceAuthorityListener: this watches for SliceAuthority calls and fills in SfaModel using the info in them
 * <p/>
 * listens to both federation API 1 and 2!
 */
public class SfaUniformFederationMemberAuthorityApiListener extends SfaModelAbstractListener {
    private static final org.slf4j.Logger LOG = LoggerFactory.getLogger(SfaUniformFederationMemberAuthorityApiListener.class);

    private GeniUserProvider geniUserProvider;

    public SfaUniformFederationMemberAuthorityApiListener(SfaModel sfaModel, GeniUserProvider geniUserProvider) {
        super(sfaModel);
        this.geniUserProvider = geniUserProvider;
    }

    private void onGetCredentials(ApiCallDetails result) {
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            GeniUrn memberUrn = (GeniUrn) result.getMethodParameters().get("memberUrn");
            assert memberUrn != null;

            List<AnyCredential> userCredentials = (List<AnyCredential>) result.getReply().getValue();
            if (userCredentials != null && userCredentials.size() > 0) {
                if (geniUserProvider != null &&
                        geniUserProvider.isUserLoggedIn() &&
                        memberUrn.equals(geniUserProvider.getLoggedInGeniUser().getUserUrn()))
                    //store first in list //TODO support multiple
                    getModel().setUserCredentials(userCredentials);
                else {
                    LOG.warn("Credentials are not for user, adding them only to parameterHistoryModel.");
                    ParameterHistoryModel parameterHistoryModel = getModel().getParameterHistoryModel();
                    userCredentials.forEach(parameterHistoryModel::addUserCredential);
                }
                if (userCredentials.size() > 1)
                    LOG.warn("TODO: The Federation API MA returned multiple credentials, but this is not yet supported by jFed");
            }
        }
    }

    private void onLookupKeysApi1(ApiCallDetails result) {
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            Map<GeniUrn, List<Map<String, Object>>> lookupResult =
                    (Map<GeniUrn, List<Map<String, Object>>>) result.getReply().getValue();

            for (GeniUrn userUrn : lookupResult.keySet()) {
                List<Map<String, Object>> userKeys = lookupResult.get(userUrn);
                for (Map<String, Object> keyInfo : userKeys) {
                    if (keyInfo.containsKey("KEY_PUBLIC")) {

                        //TODO check which member key is for     (probably possible after API has changed)
                        //  currently assumes it's the logged in users key

                        String pubKey = (String) keyInfo.get("KEY_PUBLIC");

//                        //if key is for logged in user, store it
//                        if (geniUserProvider != null &&
//                                geniUserProvider.isUserLoggedIn() &&
//                                userUrn.equals(geniUserProvider.getLoggedInGeniUser().getUserUrn()))
//                            getModel().addUserKey((String) pubKey);

                        //log any seen user key
                        getModel().addUserKey(userUrn, pubKey);
                    }
                }
            }
        }
    }

    private void onLookupKeysApi2(ApiCallDetails result) {
        if (result.getReply().getGeniResponseCode().isSuccess()) {
            AbstractFederationApi.LookupResult lookupResult =
                    (AbstractFederationApi.LookupResult) result.getReply().getValue();

            List<String> keyIds = lookupResult.getKeys();
            for (String keyId : keyIds) {
                Map<String, Object> keyInfo = lookupResult.get(keyId);
                if (keyInfo.containsKey("KEY_PUBLIC") &&
                        keyInfo.containsKey("KEY_MEMBER")) {

                    String memberUrnStr = (String) keyInfo.get("KEY_MEMBER");
                    String pubKey = (String) keyInfo.get("KEY_PUBLIC");

//                    //if key is for logged in user, store it
//                    if (geniUserProvider != null &&
//                            geniUserProvider.isUserLoggedIn() &&
//                            memberUrnStr.equals(geniUserProvider.getLoggedInGeniUser().getUserUrnString()))
//                        getModel().addUserKey((String) pubKey);

                    //log any seen user key
                    GeniUrn memberUrn = GeniUrn.parse(memberUrnStr);
                    if (memberUrn != null)
                        getModel().addUserKey(memberUrn, pubKey);
                }
            }
        }
    }

    private void onLookupMember2(ApiCallDetails result) {
        if (result.getReply().getGeniResponseCode().isSuccess() &&
                result.getReply() != null &&
                result.getReply().getValue() != null &&
                result.getReply().getValue() instanceof AbstractFederationApi.LookupResult) {
            AbstractFederationApi.LookupResult lookupResult = (AbstractFederationApi.LookupResult) result.getReply().getValue();
            String key = "_GENI_MEMBER_SSL_CERTIFICATE";
            if (lookupResult == null || lookupResult.size() != 1 || lookupResult.get(0).get(key) == null)
                return;
            String x509pemCerts = (String) lookupResult.get(0).get(key);
            List<X509Certificate> certs = KeyUtil.pemToX509CertificateChain(x509pemCerts);
            for (X509Certificate cert : certs) {
                try {
                    //a bit of a hack, to find the user certificate in the chain
                    Collection<List<?>> altNames = cert.getSubjectAlternativeNames();
                    for (List<?> altName : altNames) {
                        Integer nameType = (Integer) altName.get(0);
                        String altKey = altName.get(1).toString();

                        if (altKey.contains("+user+")) {
                            GeniUrn memberUrn = GeniUrn.parse(altKey);
                            if (memberUrn != null)
                                getModel().addUserKey(memberUrn, KeyUtil.publicKeyToOpenSshAuthorizedKeysFormat(cert.getPublicKey()));
                            break;
                        }
                    }
                } catch (CertificateParsingException e) {
                    LOG.debug("Error parsing certificate received in lookup MEMBER \"_GENI_MEMBER_SSL_CERTIFICATE\" field. Will ignore certificate.");
                    //ignore
                }
            }
        }

    }


    @Override
    public void onResult(final ApiCallDetails details) {
        assert Platform.isFxApplicationThread();
        onResultInJavaFXThread(details);
    }

    public void onResultInJavaFXThread(ApiCallDetails result) {
        //ignore errors here
        if (result.getReply() == null || result.getJavaMethodName() == null)
            return;

        if (!result.getApiName().equals(FederationMemberAuthorityApi1.getApiName()) && !result.getApiName().equals(FederationMemberAuthorityApi2.getApiName()))
            return;

        try {
            if (result.getApiName().equals(FederationMemberAuthorityApi1.getApiName())) {
                if (result.getJavaMethodName().equals("lookupKeys"))
                    onLookupKeysApi1(result);

                if (result.getJavaMethodName().equals("getCredentials"))
                    onGetCredentials(result);
            }

            if (result.getApiName().equals(FederationMemberAuthorityApi2.getApiName())) {
                if (result.getJavaMethodName().equals("lookup") &&
                        result.getMethodParameters().get("type").equals("KEY"))
                    onLookupKeysApi2(result);

                if (result.getJavaMethodName().equals("lookup") &&
                        result.getMethodParameters().get("type").equals("MEMBER"))
                    onLookupMember2(result);

                if (result.getJavaMethodName().equals("lookupMember"))
                    onLookupMember2(result);

                if (result.getJavaMethodName().equals("getCredentials"))
                    onGetCredentials(result);
            }
        } catch (Exception e) {
            LOG.warn("Exception when processing SliceAuthority reply for SfaModel. This will be ignored, but it is most likely a bug. ", e);
        }
    }
}
