package be.iminds.ilabt.jfed.lowlevel.connection;

import be.iminds.ilabt.jfed.lowlevel.JFedException;
import be.iminds.ilabt.jfed.util.*;
import org.apache.http.HttpHost;
import org.apache.http.auth.AuthScope;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.apache.http.client.AuthCache;
import org.apache.http.client.protocol.ClientContext;
import org.apache.http.conn.ClientConnectionManager;
import org.apache.http.conn.DnsResolver;
import org.apache.http.conn.scheme.Scheme;
import org.apache.http.conn.scheme.SchemeRegistry;
import org.apache.http.conn.scheme.SchemeSocketFactory;
import org.apache.http.conn.ssl.BrowserCompatHostnameVerifier;
import org.apache.http.conn.ssl.SSLSocketFactory;
import org.apache.http.conn.ssl.TrustStrategy;
import org.apache.http.impl.auth.BasicScheme;
import org.apache.http.impl.client.BasicAuthCache;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.impl.conn.PoolingClientConnectionManager;
import org.apache.http.impl.conn.SchemeRegistryFactory;
import org.apache.http.impl.conn.SystemDefaultDnsResolver;
import org.apache.http.params.*;
import org.apache.http.protocol.BasicHttpContext;
import org.apache.http.protocol.HttpContext;
import org.apache.xmlrpc.XmlRpcClient;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLException;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.UnknownHostException;
import java.security.*;
import java.security.cert.Certificate;
import java.security.cert.*;
import java.util.*;
import java.util.concurrent.TimeUnit;

/**
 * ConnectionFactory
 */
public class BasicConnectionBuilder implements ConnectionBuilder {
    private static final Logger LOG = LoggerFactory.getLogger(BasicConnectionBuilder.class);

    private ClientConnectionManager createUnsafeClientConnectionManager(DnsResolver dnsResolver){
        ClientConnectionManager unsafeCcm = null;
        try {
            SSLContext sslcontext = SSLContext.getInstance("TLS");
            sslcontext.init(null, null, null);
            SSLSocketFactory unsafeSslSocketFactory = new SSLSocketFactory(
                    sslcontext,
                    SSLSocketFactory.ALLOW_ALL_HOSTNAME_VERIFIER);
            Scheme httpsScheme = new Scheme("https", 443, unsafeSslSocketFactory);
            SchemeRegistry unsafeSchemeRegistry = new SchemeRegistry();
            unsafeSchemeRegistry.register(httpsScheme);

            return new PoolingClientConnectionManager(unsafeSchemeRegistry,
                    -1L, TimeUnit.MILLISECONDS,
                    dnsResolver);
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException("Could not create an unsafe connection manager", e);
        } catch (KeyManagementException e) {
            throw new RuntimeException("Could not create an unsafe connection manager", e);
        }
    }
    private URL serverUrl;
    private String serverUrlString;
    private JFedConnection.DebugInfo debugInfo;
    private JFedConnection.ProxyInfo proxyInfo;
    private ConnectionConfig.Authentication authentication;
    private JFedPasswordManager.LoginInfo login;
    private List<X509Certificate> clientCertChain;
    private PrivateKey privateKey;
    private ConnectionConfig.Protocol protocol;
    private JFedTrustStore conTrustStore;
    private HandleUntrustedCallback handleUntrustedCallback;
    private boolean forceSshProxyDns = false;

    private boolean hackAllowAllServerCertificates = false;
    private List<String> hackOverrideAllowedServerCertificateAlias = null; //add "*" to accept anything

    private final String userAgent;

    /**
     * class private. In fact, should only be called from BasicConnectionBuilderFactory
     */
    BasicConnectionBuilder() {
        this.userAgent = null;
    }
    BasicConnectionBuilder(String userAgent) {
        this.userAgent = userAgent;
    }

    @Override
    public void setUrl(String serverUrlString) throws MalformedURLException {
        this.serverUrlString = serverUrlString;
        this.serverUrl = new URL(serverUrlString);
    }

    @Override
    public void setUrl(URL serverUrl) {
        this.serverUrl = serverUrl;
        this.serverUrlString = serverUrl.toExternalForm();
    }

    @Override
    public void setHackAllowAllServerCertificates(boolean hackAllowAllServerCertificates) {
        this.hackAllowAllServerCertificates = hackAllowAllServerCertificates;
    }

    @Override
    public void setHackOverrideAllowedServerCertificateAlias(List<String> hackOverrideAllowedServerCertificateAlias) {
        this.hackOverrideAllowedServerCertificateAlias = hackOverrideAllowedServerCertificateAlias;
    }

    @Override
    public void setDebugInfo(JFedConnection.DebugInfo debugInfo) {
        this.debugInfo = debugInfo;
    }

    @Override
    public void setProxy(JFedConnection.ProxyInfo proxyInfo, boolean forceSshProxyDns) {
        this.proxyInfo = proxyInfo;
        this.forceSshProxyDns = forceSshProxyDns;
    }

    @Override
    public void useNoAuthentication() {
        this.authentication = ConnectionConfig.Authentication.NONE;
    }

    @Override
    public void useHttpBasicAuthentication(JFedPasswordManager.LoginInfo login) {
        this.authentication = ConnectionConfig.Authentication.HTTP_BASIC;
        this.login = login;
        if (login == null)
            throw new IllegalArgumentException("Illegal argument: login == null");
    }

    @Override
    public void useSslClientAuthentication(List<X509Certificate> clientCertChain, PrivateKey privateKey) {
        this.authentication = ConnectionConfig.Authentication.SSL_CLIENT_AUTH;
        this.clientCertChain = clientCertChain;
        this.privateKey = privateKey;
        if (privateKey == null)
            throw new IllegalArgumentException("Illegal argument: privateKey == null");
        if (clientCertChain == null)
            throw new IllegalArgumentException("Illegal argument: clientCertificateChain == null");
        if (clientCertChain.isEmpty())
            throw new IllegalArgumentException("Illegal argument: clientCertChain is empty");
    }

    @Override
    public void useHttp() {
        this.protocol = ConnectionConfig.Protocol.HTTP;
    }

    @Override
    public void useHttps(JFedTrustStore conTrustStore, HandleUntrustedCallback handleUntrustedCallback) {
        this.protocol = ConnectionConfig.Protocol.HTTPS;
        this.conTrustStore = conTrustStore;
        this.handleUntrustedCallback = handleUntrustedCallback;
        if (conTrustStore == null)
            throw new IllegalArgumentException("Illegal argument: conTrustStore == null");
    }

    private void check() {
        if (debugInfo == null) throw new RuntimeException("Configuration problem: must specify debugInfo");
        if (serverUrlString == null || serverUrl == null)
            throw new RuntimeException("Configuration problem: must specify URL");
        if (authentication == null) throw new RuntimeException("Configuration problem: must specify authentication");
        if (protocol == null) throw new RuntimeException("Configuration problem: must specify HTTP or HTTPS");
        if (protocol != ConnectionConfig.Protocol.HTTPS && authentication == ConnectionConfig.Authentication.SSL_CLIENT_AUTH)
            throw new RuntimeException("Configuration problem: incompatible options: cannot do SSL client authentication without HTTPS.");
    }


    private class HttpConParts {
        final DefaultHttpClient defaultHttpClient;
        final HttpContext httpContext;
        final HttpContext preemptiveAuthHttpContext;
        final TrustStrategy trustStrategy;

        public HttpConParts(DefaultHttpClient defaultHttpClient, HttpContext httpContext, HttpContext preemptiveAuthHttpContext, TrustStrategy trustStrategy) {
            this.defaultHttpClient = defaultHttpClient;
            this.httpContext = httpContext;
            this.preemptiveAuthHttpContext = preemptiveAuthHttpContext;
            this.trustStrategy = trustStrategy;
        }
    }

    private static class SpyingTrustStrategy implements TrustStrategy {
        private final TrustStrategy orig;
        public X509Certificate[] chain;
        private JFedConnection connection;

        public SpyingTrustStrategy(TrustStrategy orig) {
            this.orig = orig;
        }

        @Override
        public boolean isTrusted(X509Certificate[] certificate, String authType) throws CertificateException {
            if (certificate != null && certificate.length > 0) {
                chain = certificate;
                if (connection != null)
                    connection.registerServerCertficates(chain);
            }
            return orig.isTrusted(certificate, authType);
        }

        public void setConnection(JFedConnection connection) {
            LOG.debug("registering connection to SpyingTrustStrategy");
            this.connection = connection;
            if (chain != null)
                connection.registerServerCertficates(chain);
        }
    }

    private HttpConParts getConnectionHttpClientAndContext() throws JFedException {
        HttpParams params = new BasicHttpParams();
        params.setParameter(CoreConnectionPNames.SO_TIMEOUT, 120000); //read timeout in ms (default: 0 = infinite)   (long timeout, as some calls take time to finish)
        params.setParameter(CoreConnectionPNames.CONNECTION_TIMEOUT, 10000); //connection timeout in ms (default: 0 = infinite)
        if (userAgent != null) {
            //not sure which method works, so ... both
            params.setParameter(CoreProtocolPNames.USER_AGENT, userAgent);
            HttpProtocolParams.setUserAgent(params, userAgent);
            assert HttpProtocolParams.getUserAgent(params).equals(userAgent) : "Wrong user agent: "+HttpProtocolParams.getUserAgent(params);
            //update: sadly, neither of these work.
        }

        DefaultHttpClient httpClient;

        boolean useForceSshProxyDns =
                proxyInfo == null &&
                        proxyInfo instanceof JFedConnection.SshProxyInfo &&
                        this.forceSshProxyDns;

        LOG.debug("getConnectionHttpClientAndContext() with conTrustStore.getAllowedServerCertificateHostnameAliases()="+
                (conTrustStore == null ? "null conTrustStore" : conTrustStore.getAllowedServerCertificateHostnameAliases())
                +" for serverUrl="+serverUrlString);

        if (System.getProperty("proxySet") != null &&
                System.getProperty("proxySet").equals("true") &&
                System.getProperty("socksProxyHost") != null) {
            if (hackAllowAllServerCertificates) {
                LOG.error("hackAllowAllServerCertificates is not supported with socks proxy. Will not use it!");
            }
            httpClient = SocksProxyHelper.getHttpClientOverSocksProxy(params);
        } else {
            //DNS resolver needs to be stopped if this request is sent over SSH proxy.
            //because the SSH proxy will do the resolve.

            final SystemDefaultDnsResolver defaultDnsResolver = new SystemDefaultDnsResolver();
            DnsResolver dnsResolver = !useForceSshProxyDns ?
                    defaultDnsResolver :
                    new DnsResolver() {
                        //do not resolve if going over SSH proxy, let the proxy do that
                        @Override
                        public InetAddress[] resolve(String host) throws UnknownHostException {
                            JFedConnection.SshProxyInfo proxy = (JFedConnection.SshProxyInfo) proxyInfo;
                            if (!host.equals(serverUrl.getHost())) {
                                LOG.error("This hack expected resolve of \""+serverUrl.getHost()+"\" but was requested to resolve \""+host+"\". Reverting to normal resolve.");
                                return defaultDnsResolver.resolve(host);
                            }
                            //return fake resolve
                            return new InetAddress[] { InetAddress.getByAddress(host, new byte[] {127,0,0,1}) };
                        }
                    };

            if (hackAllowAllServerCertificates) {
                httpClient = new DefaultHttpClient(createUnsafeClientConnectionManager(dnsResolver), params);
            } else {
                // this causes bugs, see FEDIBBTDEV-2262:  httpClient = new DefaultHttpClient(ccmProvider.getClientConnectionManager(), params);
                //quick-and-dirty fix, see FEDIBBTDEV-2262
                PoolingClientConnectionManager poolingClientConnectionManager =
                        new PoolingClientConnectionManager(
                                SchemeRegistryFactory.createDefault(),
                                -1L, TimeUnit.MILLISECONDS,
                                dnsResolver);
                httpClient = new DefaultHttpClient(poolingClientConnectionManager, params);
            }
        }

        TrustStrategy trustStrategy = null;
        if (protocol == ConnectionConfig.Protocol.HTTPS) {
            try {
                String key_store_pass = null;
                String privatekey_store_pass =
                        "someotherpass"; //password for internal use. We don't really need one as we'll never save this keystore or share it with other objects
                KeyStore keyStore = null;
                final KeyStore trustStore = conTrustStore.getTrustStore();

                if (authentication == ConnectionConfig.Authentication.SSL_CLIENT_AUTH) {
                    key_store_pass =
                            "somepass"; //password for internal use. We don;t really need one as we'll never save this keystore.
                    keyStore = KeyStore.getInstance("JKS", "SUN");
                    keyStore.load(null, key_store_pass.toCharArray());

                    // Create CertificateChain
                    Certificate[] certs = new Certificate[clientCertChain.size()];
                    for (int i = 0; i < clientCertChain.size(); i++) {
                        X509Certificate c = clientCertChain.get(i);
                        certs[i] = c;
                        try {
                            //FEDIBBTDEV-878: check certificate validity
                            c.checkValidity(new Date());
                        } catch (CertificateExpiredException e) {
                            throw new JFedException("Certificate " + i + " (of " + clientCertChain.size() + ") in the user " +
                                    "certificate chain has expired. NotAfter=" + c.getNotAfter() + " now=" + new Date());
                        } catch (CertificateNotYetValidException e) {
                            throw new JFedException("Certificate " + i + " (of " + clientCertChain.size() + ") in the user " +
                                    "certificate chain is not yet valid. NotBefore=" + c.getNotBefore() + " now=" + new Date());
                        }
                    }

                    // storing private key and certificate to keystore
                    keyStore.setKeyEntry("authority", privateKey, privatekey_store_pass.toCharArray()/*password is used to encrypt the key inside the keystore*/, certs);

                    //add all client certificates to trust store as well as keystore!
                    for (int i = 0; i < clientCertChain.size(); i++)
                        trustStore.setCertificateEntry("clientCertificate" + i, clientCertChain.get(i));
                }

                String algo = SSLSocketFactory.TLS;
                java.security.SecureRandom random = new SecureRandom();

                final List<String> allowedCertificateHostnameAliases = hackOverrideAllowedServerCertificateAlias != null ?
                        new ArrayList<String>(hackOverrideAllowedServerCertificateAlias) :
                        new ArrayList<String>(conTrustStore.getAllowedServerCertificateHostnameAliases());

                SSLSocketFactory socketFactory;
                if (handleUntrustedCallback != null) {
                    //                socketFactory = new SSLSocketFactory(TRUSTALL, new AllowAllHostnameVerifier()); //doesn't work as we need client authentication

                    //didn't work for some reason. Probably it's missing client certificate (need to check apache source code of SSLSocketFactory to correctly write this)
                    //                socketFactory = new MyInsecureSSLAcceptAllSocketFactory(algo, keyStore, privatekey_store_pass/*password is used to encrypt the key inside the keystore*/, trustStore, random, hostnameVerifier);

                    //Solved by connecting 2 times. Not best solution, but it works.
//                    SSLCertificateDownloader.SSLCertificateJFedInfo certInfo =
//                            SSLCertificateDownloader.getCertificateInfo(serverUrl,
//                                    (proxyInfo != null && proxyInfo instanceof JFedConnection.SshProxyInfo) ? (JFedConnection.SshProxyInfo) proxyInfo : null);
//                    if (certInfo != null && certInfo.isSelfSigned() != null && certInfo.isSelfSigned()) {
//                        if (handleUntrustedCallback.trust(certInfo)) {
//                            trustStore.setCertificateEntry("allTrustCert" + extraTrustCount, certInfo.getCert());
//                            if (!certInfo.getSubjectMatchesHostname())
//                                allowedCertificateHostnameAliases.add(certInfo.getSubject());
//                        } else {
//                            LOG.info("User does not trust certificate -> Not adding anything to trust store.");
//                        }
//                    }

                    /** This is probably not a good implementation of the default TrustStrategy! Too bad the real default is nowhere to be found. */
                    trustStrategy = new TrustStrategy() {
                        @Override
                        public boolean isTrusted(X509Certificate[] certificates, String authType) throws CertificateException {
                            if (certificates == null || certificates.length == 0)
                                return false;
                            try {
                                for (X509Certificate cert : certificates) {
                                    String alias = trustStore.getCertificateAlias(cert);
                                    if (alias != null)
                                        return true;
                                }
                            } catch (KeyStoreException e) {
                                throw new CertificateException("problem with keystore while checking certificate", e);
                            }
                            SSLCertificateDownloader.SSLCertificateJFedInfo certInfo =
                                    new SSLCertificateDownloader.SSLCertificateJFedInfo(certificates[0/*certificates.length-1*/], serverUrl, serverUrl.getHost());
                            return handleUntrustedCallback.trust(certInfo);
                        }
                    };

                    LOG.debug("using home-made trustStrategy, this is far from perfect! Might be insecure!");
                }

                org.apache.http.conn.ssl.X509HostnameVerifier hostnameVerifier =
                        makeHostNameVerifier(allowedCertificateHostnameAliases, "serverUrl="+serverUrlString/*debugIdentifier*/);
                LOG.debug("Creating X509HostnameVerifier with" +
                        " allowedCertificateHostnameAliases=" + allowedCertificateHostnameAliases
                        +" for serverUrl="+serverUrlString);

                if (trustStrategy != null) {
                    LOG.debug("adding SpyingTrustStrategy");
                    trustStrategy = new SpyingTrustStrategy(trustStrategy);
                } else {
                    //too bad there is no default
//                    trustStrategy = new SpyingTrustStrategy(trustStrategy);
                }

                if (proxyInfo != null && proxyInfo instanceof JFedConnection.SshProxyInfo) {
                    JFedConnection.SshProxyInfo sshProxyInfo = (JFedConnection.SshProxyInfo) proxyInfo;

                    LOG.debug("Using SSH Proxy for connection: " + sshProxyInfo.getHostname());

                    int serverPort = serverUrl.getPort();
                    if (serverUrl.getPort() <= 0) serverPort = serverUrl.getDefaultPort();
                    SSLSocketFactory sslSocketFactory = new SshServerProxyHelper.SslOverSshProxySocketFactory(
                            sshProxyInfo,
                            algo,
                            keyStore,
                            privatekey_store_pass /* password is used to encrypt the key inside the keystore */,
                            trustStore,
                            random,
                            trustStrategy,
                            hostnameVerifier,
                            useForceSshProxyDns ? serverUrl.getHost() : null,
                            useForceSshProxyDns ? serverPort : null
                    );
                    socketFactory = sslSocketFactory;
                } else {
                    SSLSocketFactory sslSocketFactory = new SSLSocketFactory(
                            algo,
                            keyStore,
                            privatekey_store_pass/*password is used to encrypt the key inside the keystore*/,
                            trustStore,
                            random,
                            trustStrategy,
                            hostnameVerifier);
                    socketFactory = sslSocketFactory;
                }


                Scheme sch1 = new Scheme("https", 443, socketFactory); //port is the default port for the given protocol
                httpClient.getConnectionManager().getSchemeRegistry().register(sch1);
            } catch (Exception e) {
                throw new JFedException("Error creating SSL connection to " + serverUrlString, e);
            }
        }

        if (protocol == ConnectionConfig.Protocol.HTTP) {
            if (proxyInfo != null && proxyInfo instanceof JFedConnection.SshProxyInfo) {
                JFedConnection.SshProxyInfo sshProxyInfo = (JFedConnection.SshProxyInfo) proxyInfo;

                LOG.debug("Using SSH Proxy HTTP connection: " + sshProxyInfo.getHostname());

                int serverPort = serverUrl.getPort();
                if (serverUrl.getPort() <= 0) serverPort = serverUrl.getDefaultPort();
                SchemeSocketFactory socketFactory =
                        new SshServerProxyHelper.PlainHttpOverSshProxySocketFactory(sshProxyInfo,
                                useForceSshProxyDns ? serverUrl.getHost() : null,
                                useForceSshProxyDns ? serverPort : null);

                Scheme sch1 = new Scheme("http", 80/*port is the default port for the given protocol*/, socketFactory);
                httpClient.getConnectionManager().getSchemeRegistry().register(sch1);
            }
        }

        BasicHttpContext preemptiveAuthHttpContext = null;
        if (authentication == ConnectionConfig.Authentication.HTTP_BASIC) {
            int port = serverUrl.getPort();
            if (serverUrl.getPort() <= 0) port = serverUrl.getDefaultPort();
            httpClient.getCredentialsProvider().setCredentials(
                    new AuthScope(serverUrl.getHost(), port),
                    new UsernamePasswordCredentials(login.getUsername(), login.getPassword()));

            //credentials are now ready, and will be provided when server ask for it.
            //below we will construct a HttpContext that can be used to enable preemptive authentication for specific calls
            //    (= send the authentication before the server asks for it)

            //TODO: Currently, onlt HTTP / REST conenctions support this, SFA does not.

            AuthCache authCache = new BasicAuthCache();
            BasicScheme basicAuth = new BasicScheme();
            authCache.put(new HttpHost(serverUrl.getHost(), port,
                    protocol == ConnectionConfig.Protocol.HTTP ? "http" : "https"), basicAuth);
            preemptiveAuthHttpContext = new BasicHttpContext();
            preemptiveAuthHttpContext.setAttribute(ClientContext.AUTH_CACHE, authCache);
        }

        return new HttpConParts(httpClient, null, preemptiveAuthHttpContext, trustStrategy);
    }

    private static int extraTrustCount = 7000;

    public static class HostnameAndCertificateMismatchSSLException extends SSLException {
        private final java.lang.String host;
        private final String[] cns;
        private final String[] subjectAlts;
        private final List<String> allowedCertificateHostnameAliases;
        private final String origMessage;
        private final String debugIdentifier;

        private static String makeMessage(java.lang.String host, String[] cns, String[] subjectAlts, List<String> allowedCertificateHostnameAliases, String origMessage, String debugIdentifier) {
            String newMessage = origMessage == null ? "" : origMessage;
            newMessage += " (details:";
            newMessage += " host=\""+host+"\"";
            newMessage += " cns="+(cns == null ? "null" : Arrays.toString(cns));
            newMessage += " subjectAlts="+(subjectAlts == null ? "null" : Arrays.toString(subjectAlts));
            newMessage += " allowedCertificateHostnameAliases="+(allowedCertificateHostnameAliases == null ? "null" : allowedCertificateHostnameAliases);
            if (debugIdentifier != null)
                newMessage += " debugIdentifier="+debugIdentifier;
            newMessage += ")";
            return newMessage;
        }

        public HostnameAndCertificateMismatchSSLException(java.lang.String host, String[] cns, String[] subjectAlts, List<String> allowedCertificateHostnameAliases,
                                                          String origMessage, Throwable throwable, String debugIdentifier) {
            super(makeMessage(host, cns, subjectAlts, allowedCertificateHostnameAliases, origMessage, debugIdentifier), throwable);

            this.host = host;
            this.cns = cns;
            this.subjectAlts = subjectAlts;
            this.allowedCertificateHostnameAliases = allowedCertificateHostnameAliases == null ? null : new ArrayList<String>(allowedCertificateHostnameAliases);
            this.origMessage = origMessage;
            this.debugIdentifier = debugIdentifier;
        }

        public String getHost() {
            return host;
        }

        public String[] getCns() {
            return cns;
        }

        public String[] getSubjectAlts() {
            return subjectAlts;
        }

        public List<String> getAllowedCertificateHostnameAliases() {
            return allowedCertificateHostnameAliases;
        }

        public String getOrigMessage() {
            return origMessage;
        }

        public String getDebugIdentifier() {
            return debugIdentifier;
        }
    }

    private static class AliasAllowingHostnameVerifier extends org.apache.http.conn.ssl.AbstractVerifier {
        private final BrowserCompatHostnameVerifier base;
        private final List<String> allowedCertificateHostnameAliases;
        private final String debugIdentifier;

        private AliasAllowingHostnameVerifier(List<String> allowedCertificateHostnameAliases, String debugIdentifier) {
            this.allowedCertificateHostnameAliases =
                    allowedCertificateHostnameAliases == null ?
                            Collections.<String>emptyList() :
                            Collections.unmodifiableList(new ArrayList<String>(allowedCertificateHostnameAliases));
            this.base = new BrowserCompatHostnameVerifier();
            this.debugIdentifier = debugIdentifier;
        }

        /**
         * @param host the hostname of the url to check.
         * @param cns the CN's of the certificate
         * @param subjectAlts the aliases fot the CN's that the certificate allows
         * @throws javax.net.ssl.SSLException
         */
        @Override
        public final void verify(java.lang.String host, String[] cns, String[] subjectAlts) throws javax.net.ssl.SSLException {
            List<String> modifiedSubjectAlts = new ArrayList<String>();
            List<String> modifiedCns = new ArrayList<String>();


            if (subjectAlts != null)
                Collections.addAll(modifiedSubjectAlts, subjectAlts);
            if (cns != null)
                Collections.addAll(modifiedCns, cns);

            //force add of host itself, this makes any certificate we have stored match
            for (String allowedAlias : allowedCertificateHostnameAliases) {
                if (allowedAlias != null) {
                    if (modifiedCns.contains(allowedAlias) || modifiedSubjectAlts.contains(allowedAlias) || allowedAlias.equals("*")) {
                        modifiedSubjectAlts.clear();
                        modifiedCns.clear();

                        modifiedSubjectAlts.add(host);
                        modifiedCns.add(host);
                        break;
                    }
                } else {
                    LOG.error("AliasAllowingHostnameVerifier.verify -> allowedCertificateHostnameAliases contains null: "+allowedCertificateHostnameAliases+"  (this bug in the code will be safely ignored)");
                }
            }

            String[] newSubjectAlts = new String[modifiedSubjectAlts.size()];
            for (int i = 0; i < modifiedSubjectAlts.size(); i++)
                newSubjectAlts[i] = modifiedSubjectAlts.get(i);

            String[] newCns = new String[modifiedCns.size()];
            for (int i = 0; i < modifiedCns.size(); i++)
                newCns[i] = modifiedCns.get(i);

            try {
                base.verify(host, newCns, newSubjectAlts);
            } catch (javax.net.ssl.SSLException e) {
                LOG.warn("HttpsClientWithUserAuthenticationFactory hostnameVerifier host name verification failed:" + e.getMessage() + "\n" +
                        "host=" + host + "\n" +
                        "cns=" + (cns == null ? "null" : Arrays.toString(cns)) + "\n" +
                        "subjectAlts=" + Arrays.toString(subjectAlts) + "\n" +
                        "newCns=" + modifiedCns + "\n" +
                        "newSubjectAlts=" + modifiedSubjectAlts + "\n" +
                        "allowedCertificateHostnameAliases=" + allowedCertificateHostnameAliases+ "\n" +
                        "debugIdentifier=" + debugIdentifier
                        , e);
                throw new HostnameAndCertificateMismatchSSLException(
                        host,
                        cns,
                        subjectAlts,
                        allowedCertificateHostnameAliases == null ? null : new ArrayList<String>(allowedCertificateHostnameAliases),
                        e.getMessage(),
                        e,
                        debugIdentifier);
            }
        }
    }

    static org.apache.http.conn.ssl.X509HostnameVerifier makeHostNameVerifier(List<String> allowedCertificateHostnameAliases, String debugIdentifier) {
        return new AliasAllowingHostnameVerifier(allowedCertificateHostnameAliases, debugIdentifier);
    }

    @Override
    public SfaConnection buildSfaConnection() throws JFedException {
        check();

        HttpConParts httpConParts = getConnectionHttpClientAndContext();
        DefaultHttpClient httpClient = httpConParts.defaultHttpClient;

        CommonsHttpClientXmlRpcTransportFactory xmlRpcTransportFactory =
                new CommonsHttpClientXmlRpcTransportFactory(serverUrlString, httpClient);

        ConnectionConfig connectionConfig = getCurrentConnectionConfig(ConnectionConfig.Type.SFA);

        XmlRpcClient xmlRpcClient;
        try {
            xmlRpcClient = new XmlRpcClient(serverUrl, xmlRpcTransportFactory);
            BasicSfaConnection res = new BasicSfaConnection(connectionConfig, xmlRpcTransportFactory, xmlRpcClient);

            if (httpConParts.trustStrategy != null && httpConParts.trustStrategy instanceof SpyingTrustStrategy) {
                ((SpyingTrustStrategy) httpConParts.trustStrategy).setConnection(res);
            }
            return res;
        } catch (Exception e) {
            throw new JFedException("Error creating XmlRpcClient: " + e.getMessage(), e);
        }
    }

    @Override
    public HttpConnection buildHttpConnection() throws JFedException {
        check();

        HttpConParts httpConParts = getConnectionHttpClientAndContext();

        ConnectionConfig connectionConfig = getCurrentConnectionConfig(ConnectionConfig.Type.HTTP);
        BasicHttpConnection res = new BasicHttpConnection(connectionConfig,
                httpConParts.defaultHttpClient, httpConParts.httpContext, httpConParts.preemptiveAuthHttpContext);

        if (httpConParts.trustStrategy != null && httpConParts.trustStrategy instanceof SpyingTrustStrategy) {
            ((SpyingTrustStrategy) httpConParts.trustStrategy).setConnection(res);
        }
        return res;
    }

    @Override
    public ConnectionConfig getCurrentConnectionConfig(ConnectionConfig.Type type) {
        return new ConnectionConfig(serverUrl, debugInfo, proxyInfo, authentication, login, clientCertChain, privateKey, protocol, conTrustStore, handleUntrustedCallback, type);
    }

    @Override
    public void set(ConnectionConfig config) {
        this.serverUrlString = config.getServerUrlString();
        this.debugInfo = config.getDebugInfo();
        this.proxyInfo = config.getProxyInfo();
        this.authentication = config.getAuthentication();
        this.login = config.getLogin();
        this.clientCertChain = config.getClientCertChain();
        this.privateKey = config.getPrivateKey();
        this.protocol = config.getProtocol();
        this.conTrustStore = config.getConTrustStore();
        this.handleUntrustedCallback = config.getHandleUntrustedCallback();
    }


    public static class INSECURE_TRUSTALL_HandleUntrustedCallback implements HandleUntrustedCallback {
        public INSECURE_TRUSTALL_HandleUntrustedCallback() {
            LOG.warn("SECURITY WARNING: constructing INSECURE_TRUSTALL_HandleUntrustedCallback");
        }

        @Override
        public boolean trust(SSLCertificateDownloader.SSLCertificateJFedInfo sSLCertificateJFedInfo) {
            return true;
        }
    }


}
